@@ -34,7 +34,7 @@
 #include <netinet/in.h>
 #include <arpa/inet.h>
 
-#include "bscconfig.h"
+#include "config.h"
 
 #include <osmocom/core/talloc.h>
 #include <osmocom/core/select.h>
@@ -45,9 +45,9 @@
 #include <osmocom/sgsn/signal.h>
 #include <osmocom/sgsn/debug.h>
 #include <osmocom/sgsn/sgsn.h>
-#include <osmocom/sgsn/gprs_gb.h>
+#include <osmocom/sgsn/gprs_ns.h>
 #include <osmocom/sgsn/gprs_llc.h>
-#include <osmocom/sgsn/gprs_sgsn.h>
+#include <osmocom/sgsn/mmctx.h>
 #include <osmocom/sgsn/gprs_gmm.h>
 #include <osmocom/sgsn/gprs_sm.h>
 #include <osmocom/sgsn/gprs_subscriber.h>
@@ -55,8 +55,11 @@
 #include <osmocom/sgsn/gprs_ranap.h>
 #include <osmocom/sgsn/gprs_gmm_fsm.h>
 #include <osmocom/sgsn/gprs_mm_state_gb_fsm.h>
+#include <osmocom/sgsn/gtp_ggsn.h>
 #include <osmocom/sgsn/gtp_mme.h>
 #include <osmocom/sgsn/sgsn_rim.h>
+#include <osmocom/sgsn/gprs_bssgp.h>
+#include <osmocom/sgsn/pdpctx.h>
 
 #include <gtp.h>
 #include <pdp.h>
@@ -228,18 +231,11 @@ struct sgsn_pdp_ctx *sgsn_create_pdp_ctx(struct sgsn_ggsn_ctx *ggsn,
 		qos = TLVP_VAL(tp, OSMO_IE_GSM_REQ_QOS);
 	}
 
-	if (qos_len <= 3) {
-		pdp->qos_req.l = qos_len + 1;
-		if (pdp->qos_req.l > sizeof(pdp->qos_req.v))
-			pdp->qos_req.l = sizeof(pdp->qos_req.v);
-		pdp->qos_req.v[0] = 0; /* Allocation/Retention policy */
-		memcpy(&pdp->qos_req.v[1], qos, pdp->qos_req.l - 1);
-	} else {
-		pdp->qos_req.l = qos_len;
-		if (pdp->qos_req.l > sizeof(pdp->qos_req.v))
-			pdp->qos_req.l = sizeof(pdp->qos_req.v);
-		memcpy(pdp->qos_req.v, qos, pdp->qos_req.l);
-	}
+	pdp->qos_req.l = qos_len + 1;
+	if (pdp->qos_req.l > sizeof(pdp->qos_req.v))
+		pdp->qos_req.l = sizeof(pdp->qos_req.v);
+	pdp->qos_req.v[0] = 0; /* Allocation/Retention policy */
+	memcpy(&pdp->qos_req.v[1], qos, pdp->qos_req.l - 1);
 
 	/* charging characteristics if present */
 	if (TLVP_LEN(tp, OSMO_IE_GSM_CHARG_CHAR) >= sizeof(pdp->cch_pdp))
@@ -600,12 +596,12 @@ static int echo_conf(void *cbp, bool timeout)
 }
 
 /* Any message received by GGSN contains a recovery IE */
-static int cb_recovery2(struct sockaddr_in *peer, struct pdp_t *pdp, uint8_t recovery)
+static int cb_recovery3(struct gsn_t *gsn, struct sockaddr_in *peer, struct pdp_t *pdp, uint8_t recovery)
 {
 	struct sgsn_ggsn_ctx *ggsn;
 	struct sgsn_pdp_ctx *pctx = NULL;
 
-	ggsn = sgsn_ggsn_ctx_by_addr(&peer->sin_addr);
+	ggsn = sgsn_ggsn_ctx_by_addr(sgsn, &peer->sin_addr);
 	if (!ggsn) {
 		LOGP(DGPRS, LOGL_NOTICE, "Received Recovery IE for unknown GGSN\n");
 		return -EINVAL;
@@ -775,15 +771,24 @@ static int cb_data_ind(struct pdp_t *lib, void *packet, unsigned int len)
 		msgb_free(msg);
 		return -1;
 	case ST_GMM_REGISTERED_NORMAL:
-		OSMO_ASSERT(mm->gb.mm_state_fsm->state != ST_MM_IDLE);
-		if (mm->gb.mm_state_fsm->state == ST_MM_STANDBY) {
+		switch (mm->gb.mm_state_fsm->state) {
+		case ST_MM_IDLE:
+			LOGP(DGPRS, LOGL_ERROR, "Dropping DL packet for MS in MM state %s\n",
+			     osmo_fsm_inst_state_name(mm->gb.mm_state_fsm));
+			msgb_free(msg);
+			return -1;
+		case ST_MM_READY:
+			/* Go ahead */
+			break;
+		case ST_MM_STANDBY:
 			LOGMMCTXP(LOGL_INFO, mm, "Paging MS in GMM state %s, MM state %s\n",
 				  osmo_fsm_inst_state_name(mm->gmm_fsm),
 				  osmo_fsm_inst_state_name(mm->gb.mm_state_fsm));
-			gprs_gb_page_ps_ra(mm);
-		}
+			sgsn_bssgp_page_ps_ra(mm);
 
-		/* FIXME: queue the packet we received from GTP */
+			/* FIXME: queue the packet we received from GTP */
+			break;
+		}
 		break;
 	default:
 		LOGP(DGPRS, LOGL_ERROR, "GTP DATA IND for TLLI %08X in state "
@@ -800,12 +805,12 @@ static int cb_data_ind(struct pdp_t *lib, void *packet, unsigned int len)
 	/* It is easier to have a global count */
 	pdp->cdr_bytes_out += len;
 
-	return sndcp_unitdata_req(msg, &mm->gb.llme->lle[pdp->sapi],
+	return sndcp_sn_unitdata_req(msg, &mm->gb.llme->lle[pdp->sapi],
 				  pdp->nsapi, mm);
 }
 
 /* Called by SNDCP when it has received/re-assembled a N-PDU */
-int sgsn_rx_sndcp_ud_ind(struct gprs_ra_id *ra_id, int32_t tlli, uint8_t nsapi,
+int sgsn_gtp_data_req(struct gprs_ra_id *ra_id, int32_t tlli, uint8_t nsapi,
 			 struct msgb *msg, uint32_t npdu_len, uint8_t *npdu)
 {
 	struct sgsn_mm_ctx *mmctx;
@@ -908,7 +913,7 @@ int sgsn_gtp_init(struct sgsn_instance *sgi)
 	/* Register callbackcs with libgtp */
 	gtp_set_cb_delete_context(gsn, cb_delete_context);
 	gtp_set_cb_conf(gsn, cb_conf);
-	gtp_set_cb_recovery2(gsn, cb_recovery2);
+	gtp_set_cb_recovery3(gsn, cb_recovery3);
 	gtp_set_cb_data_ind(gsn, cb_data_ind);
 	gtp_set_cb_unsup_ind(gsn, cb_unsup_ind);
 	gtp_set_cb_extheader_ind(gsn, cb_extheader_ind);
