@@ -59,17 +59,18 @@
 #include <osmocom/sgsn/vty.h>
 #include <osmocom/sgsn/sgsn.h>
 #include <osmocom/sgsn/gprs_llc.h>
+#include <osmocom/sgsn/gprs_sndcp.h>
 #include <osmocom/sgsn/gprs_gmm.h>
 #include <osmocom/sgsn/gprs_ranap.h>
-#include <osmocom/sgsn/gprs_gb.h>
-
-#include <osmocom/ctrl/control_if.h>
-#include <osmocom/ctrl/ports.h>
+#include <osmocom/sgsn/gprs_ns.h>
+#include <osmocom/sgsn/gprs_bssgp.h>
+#include <osmocom/sgsn/gprs_subscriber.h>
+#include <osmocom/sgsn/gtp.h>
 
 #include <gtp.h>
 #include <osmocom/sgsn/sgsn_rim.h>
 
-#include "../../bscconfig.h"
+#include "../../config.h"
 
 #if BUILD_IU
 #include <osmocom/sigtran/osmo_ss7.h>
@@ -81,7 +82,6 @@
 #include <getopt.h>
 
 void *tall_sgsn_ctx;
-struct ctrl_handle *g_ctrlh;
 
 struct gprs_ns2_inst *sgsn_nsi;
 static int daemonize = 0;
@@ -97,34 +97,11 @@ const char *openbsc_copyright =
 
 struct sgsn_instance *sgsn;
 
-/* call-back function for the BSSGP protocol */
+/* call-back function for the BSSGP protocol.
+ * Must be left here so that we can add a new one in tests/sgsn_test */
 int bssgp_prim_cb(struct osmo_prim_hdr *oph, void *ctx)
 {
-	struct osmo_bssgp_prim *bp;
-	bp = container_of(oph, struct osmo_bssgp_prim, oph);
-
-	switch (oph->sap) {
-	case SAP_BSSGP_LL:
-		switch (oph->primitive) {
-		case PRIM_BSSGP_UL_UD:
-			return gprs_llc_rcvmsg(oph->msg, bp->tp);
-		}
-		break;
-	case SAP_BSSGP_GMM:
-		switch (oph->primitive) {
-		case PRIM_BSSGP_GMM_SUSPEND:
-			return gprs_gmm_rx_suspend(bp->ra_id, bp->tlli);
-		case PRIM_BSSGP_GMM_RESUME:
-			return gprs_gmm_rx_resume(bp->ra_id, bp->tlli,
-						  bp->u.resume.suspend_ref);
-		}
-		break;
-	case SAP_BSSGP_NM:
-		break;
-	case SAP_BSSGP_RIM:
-		return sgsn_rim_rx(bp, oph->msg);
-	}
-	return 0;
+	return sgsn_bssgp_rx_prim(oph);
 }
 
 static void signal_handler(int signum)
@@ -162,15 +139,36 @@ static void signal_handler(int signum)
 	}
 }
 
+static int sgsn_vty_go_parent(struct vty *vty)
+{
+	switch (vty->node) {
+	case SGSN_NODE:
+		vty->node = CONFIG_NODE;
+		break;
+	case MME_NODE:
+		vty->node = SGSN_NODE;
+		vty->index = NULL;
+		break;
+	default:
+#if BUILD_IU
+		osmo_ss7_vty_go_parent(vty);
+#else
+		vty->node = CONFIG_NODE;
+		vty->index = NULL;
+#endif
+		break;
+	}
+
+	return vty->node;
+}
+
 /* NSI that BSSGP uses when transmitting on NS */
 extern struct gprs_ns_inst *bssgp_nsi;
 
 static struct vty_app_info vty_info = {
 	.name 		= "OsmoSGSN",
 	.version	= PACKAGE_VERSION,
-#if BUILD_IU
-	.go_parent_cb	= osmo_ss7_vty_go_parent,
-#endif
+	.go_parent_cb	= sgsn_vty_go_parent,
 };
 
 static void print_help(void)
@@ -310,11 +308,6 @@ static struct log_info_cat gprs_categories[] = {
 		.description = "GPRS Packet Service",
 		.enabled = 1, .loglevel = LOGL_NOTICE,
 	},
-	[DNS] = {
-		.name = "DNS",
-		.description = "GPRS Network Service (NS)",
-		.enabled = 1, .loglevel = LOGL_NOTICE,
-	},
 	[DLLC] = {
 		.name = "DLLC",
 		.description = "GPRS Logical Link Control Protocol (LLC)",
@@ -410,6 +403,7 @@ int main(int argc, char **argv)
 	rate_ctr_init(tall_sgsn_ctx);
 
 	logging_vty_add_deprecated_subsys(tall_sgsn_ctx, "bssgp");
+	logging_vty_add_deprecated_subsys(tall_sgsn_ctx, "ns");
 
 	sgsn_nsi = gprs_ns2_instantiate(tall_sgsn_ctx, &gprs_ns_prim_cb, NULL);
 	if (!sgsn_nsi) {
@@ -417,19 +411,12 @@ int main(int argc, char **argv)
 		exit(1);
 	}
 	sgsn->cfg.nsi = sgsn_nsi;
-	bssgp_set_bssgp_callback(gprs_gb_send_cb, sgsn_nsi);
-
-	gprs_llc_init("/usr/local/lib/osmocom/crypt/");
-	sgsn_rate_ctr_init();
-	sgsn_inst_init(sgsn);
-
+	bssgp_set_bssgp_callback(sgsn_bssgp_dispatch_ns_unitdata_req_cb, sgsn_nsi);
 
 	gprs_ns2_vty_init(sgsn_nsi);
 	bssgp_vty_init();
 	gprs_llc_vty_init();
 	gprs_sndcp_vty_init();
-	sgsn_auth_init(sgsn);
-	sgsn_cdr_init(sgsn);
 
 	handle_options(argc, argv);
 
@@ -460,20 +447,7 @@ int main(int argc, char **argv)
 	if (rc < 0)
 		exit(1);
 
-	/* start control interface after reading config for
-	 * ctrl_vty_get_bind_addr() */
-	g_ctrlh = ctrl_interface_setup_dynip(NULL, ctrl_vty_get_bind_addr(),
-				    OSMO_CTRL_PORT_SGSN, NULL);
-	if (!g_ctrlh) {
-		LOGP(DGPRS, LOGL_ERROR, "Failed to create CTRL interface.\n");
-		exit(1);
-	}
-
-	if (sgsn_ctrl_cmds_install() != 0) {
-		LOGP(DGPRS, LOGL_ERROR, "Failed to install CTRL commands.\n");
-		exit(1);
-	}
-
+	gprs_llc_init(sgsn->cfg.crypt_cipher_plugin_path);
 
 	rc = sgsn_gtp_init(sgsn);
 	if (rc) {
@@ -482,9 +456,9 @@ int main(int argc, char **argv)
 	} else
 		LOGP(DGPRS, LOGL_NOTICE, "libGTP v%s initialized\n", gtp_version());
 
-	rc = gprs_subscr_init(sgsn);
+	rc = sgsn_inst_init(sgsn);
 	if (rc < 0) {
-		LOGP(DGPRS, LOGL_FATAL, "Cannot set up subscriber management\n");
+		LOGP(DGPRS, LOGL_FATAL, "Cannot set up SGSN\n");
 		exit(2);
 	}
 
