@@ -35,11 +35,14 @@
 #include <osmocom/sgsn/debug.h>
 #include <osmocom/sgsn/sgsn.h>
 #include <osmocom/gprs/gprs_ns2.h>
-#include <osmocom/sgsn/gprs_gb.h>
+#include <osmocom/sgsn/gprs_ns.h>
 #include <osmocom/sgsn/gprs_gmm.h>
-#include <osmocom/sgsn/gprs_sgsn.h>
+#include <osmocom/sgsn/gprs_bssgp.h>
+#include <osmocom/sgsn/mmctx.h>
+#include <osmocom/sgsn/gtp_ggsn.h>
 #include <osmocom/sgsn/gtp_mme.h>
 #include <osmocom/sgsn/vty.h>
+#include <osmocom/sgsn/pdpctx.h>
 #include <osmocom/gsupclient/gsup_client.h>
 
 #include <osmocom/vty/tdef_vty.h>
@@ -47,6 +50,7 @@
 #include <osmocom/vty/vty.h>
 #include <osmocom/vty/misc.h>
 #include <osmocom/crypt/gprs_cipher.h>
+#include <osmocom/crypt/utran_cipher.h>
 #include <osmocom/abis/ipa.h>
 
 #include <osmocom/gprs/gprs_bssgp.h>
@@ -54,14 +58,12 @@
 #include <pdp.h>
 #include <gtp.h>
 
-#include "../../bscconfig.h"
+#include "../../config.h"
 
 #ifdef BUILD_IU
 #include <osmocom/ranap/iu_client.h>
 #endif
 
-extern void *tall_sgsn_ctx;
-
 static struct sgsn_config *g_cfg = NULL;
 
 const struct value_string sgsn_auth_pol_strs[] = {
@@ -96,7 +98,7 @@ const struct value_string sgsn_auth_pol_strs[] = {
 #define NONSPEC_X1001_SECS     5       /* wait for a RANAP Release Complete */
 
 
-static struct osmo_tdef sgsn_T_defs[] = {
+struct osmo_tdef sgsn_T_defs[] = {
 	{ .T=3312, .default_val=GSM0408_T3312_SECS, .desc="Periodic RA Update timer (s)" },
 	{ .T=3313, .default_val=GSM0408_T3313_SECS, .desc="Waiting for paging response timer (s)" },
 	{ .T=3314, .default_val=GSM0408_T3314_SECS, .desc="READY timer. Force to STANDBY on expiry timer (s)" },
@@ -135,6 +137,26 @@ DEFUN(cfg_sgsn_timer, cfg_sgsn_timer_cmd,
 	return osmo_tdef_vty_set_cmd(vty, g_cfg->T_defs, argv);
 }
 
+DEFUN(show_timer_gtp, show_timer_gtp_cmd,
+      "show timer gtp " OSMO_TDEF_VTY_ARG_T_OPTIONAL,
+      SHOW_STR "Show timers\n" "GTP (libgtp) timers\n"
+      OSMO_TDEF_VTY_DOC_T)
+{
+	const char *T_arg = argc > 0 ? argv[0] : NULL;
+	return osmo_tdef_vty_show_cmd(vty, g_cfg->T_defs_gtp, T_arg, NULL);
+}
+
+DEFUN(cfg_sgsn_timer_gtp, cfg_sgsn_timer_gtp_cmd,
+      "timer gtp " OSMO_TDEF_VTY_ARG_SET_OPTIONAL,
+      "Configure or show timers\n" "GTP (libgtp) timers\n"
+      OSMO_TDEF_VTY_DOC_SET)
+{
+	/* If any arguments are missing, redirect to 'show' */
+	if (argc < 2)
+		return show_timer(self, vty, argc, argv);
+	return osmo_tdef_vty_set_cmd(vty, g_cfg->T_defs_gtp, argv);
+}
+
 char *gprs_pdpaddr2str(uint8_t *pdpa, uint8_t len, bool return_ipv6)
 {
 	static char str[INET6_ADDRSTRLEN + 10];
@@ -229,7 +251,7 @@ static int config_write_sgsn(struct vty *vty)
 	vty_out(vty, " gtp local-ip %s%s",
 		inet_ntoa(g_cfg->gtp_listenaddr.sin_addr), VTY_NEWLINE);
 
-	llist_for_each_entry(gctx, &sgsn_ggsn_ctxts, list) {
+	llist_for_each_entry(gctx, &sgsn->ggsn_list, list) {
 		if (gctx->id == UINT32_MAX)
 			continue;
 
@@ -251,15 +273,17 @@ static int config_write_sgsn(struct vty *vty)
 	for (server = sgsn->ares_servers; server; server = server->next)
 		vty_out(vty, " grx-dns-add %s%s", inet_ntoa(server->addr.addr4), VTY_NEWLINE);
 
-	if (g_cfg->cipher_support_mask != 0) {
+	if (g_cfg->gea_encryption_mask != 0) {
 		vty_out(vty, " encryption gea");
 
 		for (i = 0; i < _GPRS_ALGO_NUM; i++)
-			if (g_cfg->cipher_support_mask >> i & 1)
+			if (g_cfg->gea_encryption_mask >> i & 1)
 				vty_out(vty, " %u", i);
 
 		vty_out(vty, "%s", VTY_NEWLINE);
 	}
+	if (g_cfg->crypt_cipher_plugin_path)
+		vty_out(vty, "encryption cipher-plugin-path %s%s", g_cfg->crypt_cipher_plugin_path, VTY_NEWLINE);
 	if (g_cfg->sgsn_ipa_name)
 		vty_out(vty, " gsup ipa-name %s%s", g_cfg->sgsn_ipa_name, VTY_NEWLINE);
 	if (g_cfg->gsup_server_addr.sin_addr.s_addr)
@@ -288,9 +312,9 @@ static int config_write_sgsn(struct vty *vty)
 	llist_for_each_entry(acl, &g_cfg->imsi_acl, list)
 		vty_out(vty, " imsi-acl add %s%s", acl->imsi, VTY_NEWLINE);
 
-	if (llist_empty(&sgsn_apn_ctxts))
+	if (llist_empty(&sgsn->apn_list))
 		vty_out(vty, " ! apn * ggsn 0%s", VTY_NEWLINE);
-	llist_for_each_entry(actx, &sgsn_apn_ctxts, list) {
+	llist_for_each_entry(actx, &sgsn->apn_list, list) {
 		if (strlen(actx->imsi_prefix) > 0)
 			vty_out(vty, " apn %s imsi-prefix %s ggsn %u%s",
 				actx->name, actx->imsi_prefix, actx->ggsn->id,
@@ -311,6 +335,7 @@ static int config_write_sgsn(struct vty *vty)
 	vty_out(vty, " cdr interval %d%s", g_cfg->cdr.interval, VTY_NEWLINE);
 
 	osmo_tdef_vty_write(vty, g_cfg->T_defs, " timer ");
+	osmo_tdef_vty_write(vty, g_cfg->T_defs_gtp, " timer gtp ");
 
 	if (g_cfg->pcomp_rfc1144.active) {
 		vty_out(vty, " compression rfc1144 active slots %d%s",
@@ -398,7 +423,7 @@ DEFUN(cfg_ggsn_remote_ip, cfg_ggsn_remote_ip_cmd,
 	"IPv4 Address\n")
 {
 	uint32_t id = atoi(argv[0]);
-	struct sgsn_ggsn_ctx *ggc = sgsn_ggsn_ctx_find_alloc(id);
+	struct sgsn_ggsn_ctx *ggc = sgsn_ggsn_ctx_find_alloc(sgsn, id);
 
 	inet_aton(argv[1], &ggc->remote_addr);
 
@@ -423,7 +448,7 @@ DEFUN(cfg_ggsn_gtp_version, cfg_ggsn_gtp_version_cmd,
 	"Version 0\n" "Version 1\n")
 {
 	uint32_t id = atoi(argv[0]);
-	struct sgsn_ggsn_ctx *ggc = sgsn_ggsn_ctx_find_alloc(id);
+	struct sgsn_ggsn_ctx *ggc = sgsn_ggsn_ctx_find_alloc(sgsn, id);
 
 	if (atoi(argv[1]))
 		ggc->gtp_version = 1;
@@ -441,7 +466,7 @@ DEFUN(cfg_ggsn_echo_interval, cfg_ggsn_echo_interval_cmd,
 	"Interval between echo requests in seconds.\n")
 {
 	uint32_t id = atoi(argv[0]);
-	struct sgsn_ggsn_ctx *ggc = sgsn_ggsn_ctx_find_alloc(id);
+	struct sgsn_ggsn_ctx *ggc = sgsn_ggsn_ctx_find_alloc(sgsn, id);
 
 	ggc->echo_interval = atoi(argv[1]);
 
@@ -460,7 +485,7 @@ DEFUN(cfg_ggsn_no_echo_interval, cfg_ggsn_no_echo_interval_cmd,
 	NO_STR "Send an echo request to this static GGSN every interval.\n")
 {
 	uint32_t id = atoi(argv[0]);
-	struct sgsn_ggsn_ctx *ggc = sgsn_ggsn_ctx_find_alloc(id);
+	struct sgsn_ggsn_ctx *ggc = sgsn_ggsn_ctx_find_alloc(sgsn, id);
 
 	ggc->echo_interval = 0;
 	sgsn_ggsn_ctx_check_echo_timer(ggc);
@@ -501,7 +526,7 @@ static int add_apn_ggsn_mapping(struct vty *vty, const char *apn_str,
 	struct apn_ctx *actx;
 	struct sgsn_ggsn_ctx *ggsn;
 
-	ggsn = sgsn_ggsn_ctx_by_id(ggsn_id);
+	ggsn = sgsn_ggsn_ctx_by_id(sgsn, ggsn_id);
 	if (ggsn == NULL) {
 		vty_out(vty, "%% a GGSN with id %d has not been defined%s",
 			ggsn_id, VTY_NEWLINE);
@@ -688,7 +713,7 @@ DEFUN(swow_mmctx_all, show_mmctx_all_cmd,
 	SHOW_STR MMCTX_STR "All MM Contexts\n" INCLUDE_PDP_STR)
 {
 	struct sgsn_mm_ctx *mm;
-	llist_for_each_entry(mm, &sgsn_mm_ctxts, list)
+	llist_for_each_entry(mm, &sgsn->mm_list, list)
 		vty_dump_mmctx(vty, "", mm, (argc > 0) ? 1 : 0);
 
 	return CMD_SUCCESS;
@@ -700,7 +725,7 @@ DEFUN(show_pdpctx_all, show_pdpctx_all_cmd,
 {
 	struct sgsn_pdp_ctx *pdp;
 
-	llist_for_each_entry(pdp, &sgsn_pdp_ctxts, g_list)
+	llist_for_each_entry(pdp, &sgsn->pdp_list, g_list)
 		vty_dump_pdp(vty, "", pdp);
 
 	return CMD_SUCCESS;
@@ -770,14 +795,16 @@ DEFUN_DEPRECATED(cfg_encrypt, cfg_encrypt_cmd,
 		}
 	}
 
-	g_cfg->cipher_support_mask |= (1 << c);
+	g_cfg->gea_encryption_mask |= (1 << c);
 
 	return CMD_SUCCESS;
 }
 
+#define ENCRYPTION_STR "Set encryption algorithms for SGSN\n"
+
 DEFUN(cfg_encrypt2, cfg_encrypt2_cmd,
 	"encryption gea <0-4> [<0-4>] [<0-4>] [<0-4>] [<0-4>]",
-	"Set encryption algorithms for SGSN\n"
+	ENCRYPTION_STR
 	"GPRS Encryption Algorithm\n"
 	"GEAn Algorithm Number\n"
 	"GEAn Algorithm Number\n"
@@ -787,12 +814,12 @@ DEFUN(cfg_encrypt2, cfg_encrypt2_cmd,
 {
 	int i = 0;
 
-	g_cfg->cipher_support_mask = 0;
+	g_cfg->gea_encryption_mask = 0;
 	for (i = 0; i < argc; i++)
-		g_cfg->cipher_support_mask |= (1 << atoi(argv[i]));
+		g_cfg->gea_encryption_mask |= (1 << atoi(argv[i]));
 
 	for (i = 0; i < _GPRS_ALGO_NUM; i++) {
-		if (g_cfg->cipher_support_mask >> i & 1) {
+		if (g_cfg->gea_encryption_mask >> i & 1) {
 
 			if (i == GPRS_ALGO_GEA0)
 				continue;
@@ -814,6 +841,27 @@ DEFUN(cfg_encrypt2, cfg_encrypt2_cmd,
 	return CMD_SUCCESS;
 }
 
+DEFUN(cfg_encrypt_cipher_plugin_path, cfg_encrypt_cipher_plugin_path_cmd,
+	"encryption cipher-plugin-path PATH",
+	ENCRYPTION_STR
+	"Path to gprs encryption cipher plugin directory\n"
+	"Plugin path\n")
+{
+	osmo_talloc_replace_string(sgsn, &sgsn->cfg.crypt_cipher_plugin_path, argv[0]);
+
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_no_encrypt_cipher_plugin_path, cfg_no_encrypt_cipher_plugin_path_cmd,
+	"no encryption cipher-plugin-path PATH",
+	NO_STR ENCRYPTION_STR
+	"Path to gprs encryption cipher plugin directory\n"
+	"Plugin path\n")
+{
+	TALLOC_FREE(sgsn->cfg.crypt_cipher_plugin_path);
+	return CMD_SUCCESS;
+}
+
 DEFUN(cfg_authentication, cfg_authentication_cmd,
       "authentication (optional|required)",
       "Whether to enforce MS authentication in GERAN (only with auth-policy remote)\n"
@@ -835,6 +883,23 @@ DEFUN(cfg_authentication, cfg_authentication_cmd,
 	return CMD_SUCCESS;
 }
 
+DEFUN(cfg_encryption_uea, cfg_encryption_uea_cmd,
+      "encryption uea <0-2> [<0-2>] [<0-2>]",
+      ENCRYPTION_STR
+      "UTRAN (3G) encryption algorithms to allow: 0 = UEA0 (no encryption), 1 = UEA1, 2 = UEA2.\n"
+      "UEAn Algorithm Number\n"
+      "UEAn Algorithm Number\n"
+      "UEAn Algorithm Number\n")
+{
+	unsigned int i;
+
+	g_cfg->uea_encryption_mask = 0;
+	for (i = 0; i < argc; i++)
+		g_cfg->uea_encryption_mask |= (1 << atoi(argv[i]));
+
+	return CMD_SUCCESS;
+}
+
 DEFUN(cfg_auth_policy, cfg_auth_policy_cmd,
 	"auth-policy (accept-all|closed|acl-only|remote)",
 	"Configure the Authorization policy of the SGSN. This setting determines which subscribers are"
@@ -964,7 +1029,7 @@ DEFUN_HIDDEN(reset_sgsn_state,
 	struct gprs_subscr *subscr, *tmp_subscr;
 	struct sgsn_mm_ctx *mm, *tmp_mm;
 
-	llist_for_each_entry_safe(mm, tmp_mm, &sgsn_mm_ctxts, list)
+	llist_for_each_entry_safe(mm, tmp_mm, &sgsn->mm_list, list)
 	{
 		gsm0408_gprs_access_cancelled(mm, SGSN_ERROR_CAUSE_NONE);
 	}
@@ -1243,7 +1308,7 @@ DEFUN(page_subscr, page_subscr_info_cmd,
 		return CMD_WARNING;
 	}
 
-	gprs_gb_page_ps_ra(mm);
+	sgsn_bssgp_page_ps_ra(mm);
 	return CMD_SUCCESS;
 }
 
@@ -1696,9 +1761,6 @@ int sgsn_vty_init(struct sgsn_config *cfg)
 {
 	g_cfg = cfg;
 
-	g_cfg->T_defs = sgsn_T_defs;
-	osmo_tdefs_reset(g_cfg->T_defs);
-
 	install_element_ve(&show_sgsn_cmd);
 	//install_element_ve(&show_mmctx_tlli_cmd);
 	install_element_ve(&show_mmctx_imsi_cmd);
@@ -1706,6 +1768,7 @@ int sgsn_vty_init(struct sgsn_config *cfg)
 	install_element_ve(&show_pdpctx_all_cmd);
 	install_element_ve(&show_subscr_cache_cmd);
 	install_element_ve(&show_timer_cmd);
+	install_element_ve(&show_timer_gtp_cmd);
 
 	install_element(ENABLE_NODE, &update_subscr_insert_auth_triplet_cmd);
 	install_element(ENABLE_NODE, &update_subscr_create_cmd);
@@ -1732,6 +1795,9 @@ int sgsn_vty_init(struct sgsn_config *cfg)
 	/* order matters here: ensure we attempt to parse our new command first! */
 	install_element(SGSN_NODE, &cfg_encrypt2_cmd);
 	install_element(SGSN_NODE, &cfg_encrypt_cmd);
+	install_element(SGSN_NODE, &cfg_encryption_uea_cmd);
+	install_element(SGSN_NODE, &cfg_encrypt_cipher_plugin_path_cmd);
+	install_element(SGSN_NODE, &cfg_no_encrypt_cipher_plugin_path_cmd);
 
 	install_element(SGSN_NODE, &cfg_gsup_ipa_name_cmd);
 	install_element(SGSN_NODE, &cfg_gsup_remote_ip_cmd);
@@ -1752,6 +1818,7 @@ int sgsn_vty_init(struct sgsn_config *cfg)
 	install_element(SGSN_NODE, &cfg_grx_ggsn_cmd);
 
 	install_element(SGSN_NODE, &cfg_sgsn_timer_cmd);
+	install_element(SGSN_NODE, &cfg_sgsn_timer_gtp_cmd);
 
 	install_element(SGSN_NODE, &cfg_no_comp_rfc1144_cmd);
 	install_element(SGSN_NODE, &cfg_comp_rfc1144_cmd);
@@ -1783,7 +1850,8 @@ int sgsn_parse_config(const char *config_file)
 	/* make sure sgsn_vty_init() was called before this */
 	OSMO_ASSERT(g_cfg);
 
-	g_cfg->cipher_support_mask = 0x1; /* support GEA0 by default unless specific encryption config exists */
+	g_cfg->gea_encryption_mask = 0x1; /* support GEA0 by default unless specific encryption config exists */
+	g_cfg->uea_encryption_mask = (1 << OSMO_UTRAN_UEA0); /* support UEA0 by default unless specific encryption config exists */
 
 	rc = vty_read_config_file(config_file, NULL);
 	if (rc < 0) {
